@@ -397,7 +397,6 @@ add_profile_button_clicked_cb (NetDeviceEthernet *self)
         g_autofree gchar *id = NULL;
         NetConnectionEditor *editor;
         const GPtrArray *connections;
-        const char *iface;
 
         connection = nm_simple_connection_new ();
         sc = NM_SETTING_CONNECTION (nm_setting_connection_new ());
@@ -415,11 +414,6 @@ add_profile_button_clicked_cb (NetDeviceEthernet *self)
                       NM_SETTING_CONNECTION_AUTOCONNECT, TRUE,
                       NULL);
 
-        iface = nm_device_get_iface (self->device);
-        g_object_set (sc,
-                      NM_SETTING_CONNECTION_INTERFACE_NAME, iface,
-                      NULL);
-
         nm_connection_add_setting (connection, nm_setting_wired_new ());
 
         editor = net_connection_editor_new (connection, self->device, NULL, self->client);
@@ -445,7 +439,7 @@ device_off_switch_changed_cb (NetDeviceEthernet *self)
                                                              NULL, NULL, NULL, NULL);
                 }
         } else {
-                nm_device_disconnect (self->device, NULL, NULL);
+                nm_device_disconnect_async (self->device, NULL, NULL, NULL);
         }
 }
 
