@@ -347,6 +347,9 @@ als_enabled_state_changed (CcPowerPanel *self)
         visible = g_variant_get_boolean (v);
     }
 
+  if (gtk_widget_get_visible (GTK_WIDGET (self->als_row)) == visible)
+    return;
+
   enabled = g_settings_get_boolean (self->gsd_settings, "ambient-enabled");
   g_debug ("ALS enabled: %s", enabled ? "on" : "off");
   g_signal_handlers_block_by_func (self->als_switch, als_switch_changed_cb, self);
@@ -990,10 +993,7 @@ performance_profile_set_active (CcPowerPanel  *self,
   GtkCheckButton *button;
 
   button = cc_power_profile_row_get_radio_button (CC_POWER_PROFILE_ROW (self->power_profiles_row[profile]));
-  if (!button) {
-    g_warning ("Not setting profile '%s' as it doesn't have a widget", profile_str);
-    return;
-  }
+  g_assert (button);
   gtk_check_button_set_active (GTK_CHECK_BUTTON (button), TRUE);
 }
 
