@@ -31,10 +31,27 @@
 
 #include "config.h"
 
+#define CC_FPRINTD_NAME "net.reactivated.Fprint"
+
 /* Translate fprintd strings */
 #define TR(s) dgettext ("fprintd", s)
 #include "fingerprint-strings.h"
 
+typedef enum {
+  DIALOG_STATE_NONE                   = 0,
+  DIALOG_STATE_DEVICES_LISTING        = (1 << 0),
+  DIALOG_STATE_DEVICE_CLAIMING        = (1 << 1),
+  DIALOG_STATE_DEVICE_CLAIMED         = (1 << 2),
+  DIALOG_STATE_DEVICE_PRINTS_LISTING  = (1 << 3),
+  DIALOG_STATE_DEVICE_RELEASING       = (1 << 4),
+  DIALOG_STATE_DEVICE_ENROLL_STARTING = (1 << 5),
+  DIALOG_STATE_DEVICE_ENROLLING       = (1 << 6),
+  DIALOG_STATE_DEVICE_ENROLL_STOPPING = (1 << 7),
+  DIALOG_STATE_DEVICE_DELETING        = (1 << 8),
+
+  DIALOG_STATE_IDLE = DIALOG_STATE_DEVICE_CLAIMED | DIALOG_STATE_DEVICE_ENROLLING,
+} DialogState;
+
 struct _CcFingerprintDialog
 {
   GtkWindow parent_instance;
@@ -43,7 +60,7 @@ struct _CcFingerprintDialog
   GtkButton      *cancel_button;
   GtkButton      *delete_prints_button;
   GtkButton      *done_button;
-  GtkContainer   *add_print_popover_box;
+  GtkBox         *add_print_popover_box;
   GtkEntry       *enroll_print_entry;
   GtkFlowBox     *prints_gallery;
   GtkHeaderBar   *titlebar;
@@ -53,8 +70,7 @@ struct _CcFingerprintDialog
   GtkLabel       *infobar_error;
   GtkLabel       *title;
   GtkListBox     *devices_list;
-  GtkPopoverMenu *add_print_popover;
-  GtkPopoverMenu *print_popover;
+  GtkPopover     *add_print_popover;
   GtkSpinner     *spinner;
   GtkStack       *stack;
   GtkWidget      *add_print_icon;
@@ -68,14 +84,12 @@ struct _CcFingerprintDialog
   GtkWidget      *prints_manager;
 
   CcFingerprintManager *manager;
+  DialogState           dialog_state;
   CcFprintdDevice      *device;
-  gboolean              claiming;
-  gboolean              device_claimed;
   gulong                device_signal_id;
   gulong                device_name_owner_id;
   GCancellable         *cancellable;
   GStrv                 enrolled_fingers;
-  const char           *enrolling_finger;
   guint                 enroll_stages_passed;
   guint                 enroll_stage_passed_id;
   gdouble               enroll_progress;
@@ -144,6 +158,104 @@ cc_fingerprint_dialog_new (CcFingerprintManager *manager)
                        NULL);
 }
 
+static gboolean
+update_dialog_state (CcFingerprintDialog *self,
+                     DialogState         state)
+{
+  if (self->dialog_state == state)
+    return FALSE;
+
+  self->dialog_state = state;
+
+  if (self->dialog_state == DIALOG_STATE_NONE ||
+      self->dialog_state == (self->dialog_state & DIALOG_STATE_IDLE))
+    {
+      gtk_spinner_stop (self->spinner);
+    }
+  else
+    {
+      gtk_spinner_start (self->spinner);
+    }
+
+  return TRUE;
+}
+
+static gboolean
+add_dialog_state (CcFingerprintDialog *self,
+                  DialogState          state)
+{
+  return update_dialog_state (self, (self->dialog_state | state));
+}
+
+static gboolean
+remove_dialog_state (CcFingerprintDialog *self,
+                     DialogState          state)
+{
+  return update_dialog_state (self, (self->dialog_state & ~state));
+}
+
+typedef struct
+{
+  CcFingerprintDialog *dialog;
+  DialogState          state;
+} DialogStateRemover;
+
+static DialogStateRemover *
+auto_state_remover (CcFingerprintDialog *self,
+                    DialogState          state)
+{
+  DialogStateRemover *state_remover;
+
+  state_remover = g_new0 (DialogStateRemover, 1);
+  state_remover->dialog = g_object_ref (self);
+  state_remover->state = state;
+
+  return state_remover;
+}
+
+static void
+auto_state_remover_cleanup (DialogStateRemover *state_remover)
+{
+  remove_dialog_state (state_remover->dialog, state_remover->state);
+  g_clear_object (&state_remover->dialog);
+  g_free (state_remover);
+}
+
+G_DEFINE_AUTOPTR_CLEANUP_FUNC (DialogStateRemover, auto_state_remover_cleanup);
+
+static const char *
+dbus_error_to_human (CcFingerprintDialog *self,
+                     GError              *error)
+{
+  g_autofree char *dbus_error = g_dbus_error_get_remote_error (error);
+
+  if (dbus_error == NULL)
+    { /* Fallback to generic */ }
+  else if (g_str_equal (dbus_error, CC_FPRINTD_NAME ".Error.ClaimDevice"))
+    return _("the device needs to be claimed to perform this action");
+  else if (g_str_equal (dbus_error, CC_FPRINTD_NAME ".Error.AlreadyInUse"))
+    return _("the device is already claimed by another process");
+  else if (g_str_equal (dbus_error, CC_FPRINTD_NAME ".Error.PermissionDenied"))
+    return _("you do not have permission to perform the action");
+  else if (g_str_equal (dbus_error, CC_FPRINTD_NAME ".Error.NoEnrolledPrints"))
+    return _("no prints have been enrolled");
+  else if (g_str_equal (dbus_error, CC_FPRINTD_NAME ".Error.NoActionInProgress"))
+    { /* Fallback to generic */ }
+  else if (g_str_equal (dbus_error, CC_FPRINTD_NAME ".Error.InvalidFingername"))
+    { /* Fallback to generic */ }
+  else if (g_str_equal (dbus_error, CC_FPRINTD_NAME ".Error.Internal"))
+    { /* Fallback to generic */ }
+
+  if (self->dialog_state & DIALOG_STATE_DEVICE_ENROLLING)
+    return _("Failed to communicate with the device during enrollment");
+
+  if (self->dialog_state & DIALOG_STATE_DEVICE_CLAIMED ||
+      self->dialog_state & DIALOG_STATE_DEVICE_CLAIMING)
+    return _("Failed to communicate with the fingerprint reader");
+
+  return _("Failed to communicate with the fingerprint daemon");
+}
+
 static void
 disconnect_device_signals (CcFingerprintDialog *self)
 {
@@ -170,11 +282,11 @@ cc_fingerprint_dialog_dispose (GObject *object)
 
   g_clear_handle_id (&self->enroll_stage_passed_id, g_source_remove);
 
-  if (self->device && self->device_claimed)
+  if (self->device && (self->dialog_state & DIALOG_STATE_DEVICE_CLAIMED))
     {
       disconnect_device_signals (self);
 
-      if (self->enrolling_finger)
+      if (self->dialog_state & DIALOG_STATE_DEVICE_ENROLLING)
         cc_fprintd_device_call_enroll_stop_sync (self->device, NULL, NULL);
       cc_fprintd_device_call_release (self->device, NULL, NULL, NULL);
     }
@@ -237,64 +349,6 @@ notify_error (CcFingerprintDialog *self,
   gtk_widget_set_visible (self->error_infobar, error_message != NULL);
 }
 
-static gboolean
-fingerprint_icon_draw (GtkWidget *widget,
-                       cairo_t   *cr,
-                       gdouble   *progress_data)
-{
-  gdouble progress = 0.0f;
-
-  if (progress_data)
-    progress = *progress_data;
-
-  if (G_APPROX_VALUE (progress, 0.f, FLT_EPSILON) || progress > 1)
-    return FALSE;
-
-  GTK_WIDGET_GET_CLASS (widget)->draw (widget, cr);
-
-  if (progress > 0)
-    {
-      g_autoptr(GdkRGBA) outline_color = NULL;
-      GtkStyleContext *context;
-      GtkStateFlags state;
-      int outline_width;
-      int outline_offset;
-      int width;
-      int height;
-      int radius;
-      int delta;
-
-      context = gtk_widget_get_style_context (widget);
-      gtk_style_context_save (context);
-
-      state = gtk_style_context_get_state (context);
-
-      gtk_style_context_add_class (context, "progress");
-      gtk_style_context_get (context, state,
-                             "outline-width", &outline_width,
-                             "outline-offset", &outline_offset,
-                             "outline-color", &outline_color,
-                             NULL);
-
-      width = gtk_widget_get_allocated_width (widget);
-      height = gtk_widget_get_allocated_height (widget);
-      radius = MIN (width / 2, height / 2) + outline_offset;
-      delta = radius - outline_width / 2;
-
-      cairo_arc (cr, width / 2., height / 2., delta,
-                 1.5 * G_PI, (1.5 + progress * 2) * G_PI);
-      gdk_cairo_set_source_rgba (cr, outline_color);
-
-      cairo_set_line_width (cr, MIN (outline_width, radius));
-      cairo_set_line_cap (cr, CAIRO_LINE_CAP_BUTT);
-      cairo_stroke (cr);
-
-      gtk_style_context_restore (context);
-    }
-
-  return TRUE;
-}
-
 static GtkWidget *
 fingerprint_icon_new (const char *icon_name,
                       const char *label_text,
@@ -315,21 +369,16 @@ fingerprint_icon_new (const char *icon_name,
   gtk_widget_set_name (box, "fingerprint-box");
   gtk_widget_set_hexpand (box, TRUE);
 
-  image = gtk_image_new_from_icon_name (icon_name, GTK_ICON_SIZE_DND);
+  image = gtk_image_new_from_icon_name (icon_name);
 
   if (icon_widget_type == GTK_TYPE_IMAGE)
     icon_widget = image;
   else
     icon_widget = g_object_new (icon_widget_type, NULL);
 
-  if (progress_data)
-    g_signal_connect (image, "draw", G_CALLBACK (fingerprint_icon_draw),
-                      progress_data);
-
-  if (g_type_is_a (icon_widget_type, GTK_TYPE_BUTTON))
+  if (g_type_is_a (icon_widget_type, GTK_TYPE_MENU_BUTTON))
     {
-      gtk_button_set_image (GTK_BUTTON (icon_widget), image);
-      gtk_button_set_relief (GTK_BUTTON (icon_widget), GTK_RELIEF_NONE);
+      gtk_menu_button_set_child (GTK_MENU_BUTTON (icon_widget), image);
       gtk_widget_set_can_focus (icon_widget, FALSE);
     }
 
@@ -337,13 +386,13 @@ fingerprint_icon_new (const char *icon_name,
   gtk_widget_set_valign (icon_widget, GTK_ALIGN_CENTER);
   gtk_widget_set_name (icon_widget, "fingerprint-image");
 
-  gtk_container_add (GTK_CONTAINER (box), icon_widget);
+  gtk_box_append (GTK_BOX (box), icon_widget);
 
   context = gtk_widget_get_style_context (icon_widget);
-  gtk_style_context_add_class (context, "fingerprint-image");
+  gtk_style_context_add_class (context, "circular");
 
   label = gtk_label_new_with_mnemonic (label_text);
-  gtk_container_add (GTK_CONTAINER (box), label);
+  gtk_box_append (GTK_BOX (box), label);
 
   context = gtk_widget_get_style_context (box);
   gtk_style_context_add_class (context, "fingerprint-icon");
@@ -373,11 +422,11 @@ fingerprint_menu_button (const char *icon_name,
   gtk_widget_set_focus_on_click (flowbox_child, FALSE);
   gtk_widget_set_name (flowbox_child, "fingerprint-flowbox");
 
-  gtk_container_add (GTK_CONTAINER (flowbox_child), box);
+  gtk_flow_box_child_set_child (GTK_FLOW_BOX_CHILD (flowbox_child), box);
 
   g_object_set_data (G_OBJECT (flowbox_child), "button", button);
   g_object_set_data (G_OBJECT (flowbox_child), "icon",
-                     gtk_button_get_image (GTK_BUTTON (button)));
+                     GTK_IMAGE (gtk_menu_button_get_child (GTK_MENU_BUTTON (button))));
   g_object_set_data (G_OBJECT (flowbox_child), "label", label);
   g_object_set_data (G_OBJECT (button), "flowbox-child", flowbox_child);
 
@@ -405,6 +454,24 @@ prints_visibility_filter (GtkFlowBoxChild *child,
   return g_strv_contains ((const gchar **) self->enrolled_fingers, finger_id);
 }
 
+static GList *
+get_container_children (GtkWidget *container)
+{
+  GtkWidget *child;
+  GList *list = NULL;
+    
+  child = gtk_widget_get_first_child (container);
+  while (child) {
+      GtkWidget *next = gtk_widget_get_next_sibling (child);
+
+      list = g_list_append (list, child);
+
+      child = next;
+  }
+
+  return list;
+}
+
 static void
 update_prints_to_add_visibility (CcFingerprintDialog *self)
 {
@@ -412,7 +479,7 @@ update_prints_to_add_visibility (CcFingerprintDialog *self)
   GList *l;
   guint i;
 
-  print_buttons = gtk_container_get_children (self->add_print_popover_box);
+  print_buttons = get_container_children (GTK_WIDGET (self->add_print_popover_box));
 
   for (i = 0, l = print_buttons; i < N_VALID_FINGERS && l; ++i, l = l->next)
     {
@@ -442,6 +509,7 @@ list_enrolled_cb (GObject      *object,
 {
   g_auto(GStrv) enrolled_fingers = NULL;
   g_autoptr(GError) error = NULL;
+  g_autoptr(DialogStateRemover) state_remover = NULL;
   CcFprintdDevice *fprintd_device = CC_FPRINTD_DEVICE (object);
   CcFingerprintDialog *self = user_data;
   guint n_enrolled_fingers = 0;
@@ -453,23 +521,23 @@ list_enrolled_cb (GObject      *object,
   if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
     return;
 
-  gtk_spinner_stop (self->spinner);
+  state_remover = auto_state_remover (self, DIALOG_STATE_DEVICE_PRINTS_LISTING);
+
   gtk_widget_set_sensitive (GTK_WIDGET (self->add_print_icon), TRUE);
 
-  if (self->device_claimed)
+  if (self->dialog_state & DIALOG_STATE_DEVICE_CLAIMED)
     gtk_widget_set_sensitive (GTK_WIDGET (self->prints_manager), TRUE);
 
   if (error)
     {
       g_autofree char *dbus_error = g_dbus_error_get_remote_error (error);
 
-      if (!dbus_error || !g_str_has_suffix (dbus_error, ".Error.NoEnrolledPrints"))
+      if (!dbus_error || !g_str_equal (dbus_error, CC_FPRINTD_NAME ".Error.NoEnrolledPrints"))
         {
           g_autofree char *error_message = NULL;
 
-          g_dbus_error_strip_remote_error (error);
           error_message = g_strdup_printf (_("Failed to list fingerprints: %s"),
-                                           error->message);
+                                           dbus_error_to_human (self, error));
           g_warning ("Listing of fingerprints on device %s failed: %s",
                      cc_fprintd_device_get_name (self->device), error->message);
           notify_error (self, error_message);
@@ -501,7 +569,9 @@ update_prints_store (CcFingerprintDialog *self)
 
   g_assert_true (CC_FPRINTD_IS_DEVICE (self->device));
 
-  gtk_spinner_start (self->spinner);
+  if (!add_dialog_state (self, DIALOG_STATE_DEVICE_PRINTS_LISTING))
+    return;
+
   gtk_widget_set_sensitive (GTK_WIDGET (self->add_print_icon), FALSE);
   gtk_widget_hide (GTK_WIDGET (self->delete_prints_button));
 
@@ -533,9 +603,8 @@ delete_prints_cb (GObject      *object,
     {
       g_autofree char *error_message = NULL;
 
-      g_dbus_error_strip_remote_error (error);
       error_message = g_strdup_printf (_("Failed to delete saved fingerprints: %s"),
-                                       error->message);
+                                       dbus_error_to_human (self, error));
       g_warning ("Deletion of fingerprints on device %s failed: %s",
                  cc_fprintd_device_get_name (self->device), error->message);
       notify_error (self, error_message);
@@ -548,10 +617,12 @@ delete_prints_cb (GObject      *object,
 static void
 delete_enrolled_prints (CcFingerprintDialog *self)
 {
-  g_return_if_fail (self->device_claimed);
+  g_return_if_fail (self->dialog_state & DIALOG_STATE_DEVICE_CLAIMED);
+
+  if (!add_dialog_state (self, DIALOG_STATE_DEVICE_DELETING))
+    return;
 
   gtk_widget_set_sensitive (GTK_WIDGET (self->prints_manager), FALSE);
-  gtk_spinner_start (self->spinner);
 
   cc_fprintd_device_call_delete_enrolled_fingers2 (self->device,
                                                    self->cancellable,
@@ -591,7 +662,7 @@ have_multiple_devices (CcFingerprintDialog *self)
 {
   g_autoptr(GList) devices_rows = NULL;
 
-  devices_rows = gtk_container_get_children (GTK_CONTAINER (self->devices_list));
+  devices_rows = get_container_children (GTK_WIDGET (self->devices_list));
 
   return devices_rows && devices_rows->next;
 }
@@ -627,7 +698,7 @@ set_enroll_result_message (CcFingerprintDialog *self,
 
   gtk_style_context_add_class (style_context, ENROLL_STATE_CLASSES[enroll_state]);
 
-  gtk_image_set_from_icon_name (self->enroll_result_image, icon_name, GTK_ICON_SIZE_DND);
+  gtk_image_set_from_icon_name (self->enroll_result_image, icon_name);
   gtk_label_set_label (self->enroll_result_message, message);
 }
 
@@ -651,7 +722,7 @@ handle_enroll_signal (CcFingerprintDialog *self,
 {
   gboolean completed;
 
-  g_return_if_fail (self->enrolling_finger);
+  g_return_if_fail (self->dialog_state & DIALOG_STATE_DEVICE_ENROLLING);
 
   g_debug ("Device enroll result message: %s, done: %d", result, done);
 
@@ -703,7 +774,7 @@ handle_enroll_signal (CcFingerprintDialog *self,
       scan_type = cc_fprintd_device_get_scan_type (self->device);
       is_swipe = g_str_equal (scan_type, "swipe");
 
-      message = TR (enroll_result_str_to_msg (result, is_swipe));
+      message = enroll_result_str_to_msg (result, is_swipe);
       set_enroll_result_message (self, ENROLL_STATE_RETRY, message);
 
       self->enroll_stage_passed_id =
@@ -727,11 +798,19 @@ handle_enroll_signal (CcFingerprintDialog *self,
           const char *message;
 
           if (g_str_equal (result, "enroll-disconnected"))
-            message = _("Fingerprint device disconnected");
+            {
+              message = _("Fingerprint device disconnected");
+              remove_dialog_state (self, DIALOG_STATE_DEVICE_CLAIMED |
+                                         DIALOG_STATE_DEVICE_ENROLLING);
+            }
           else if (g_str_equal (result, "enroll-data-full"))
-            message = _("Fingerprint device storage is full");
+            {
+              message = _("Fingerprint device storage is full");
+            }
           else
-            message = _("Failed to enroll new fingerprint");
+            {
+              message = _("Failed to enroll new fingerprint");
+            }
 
           set_enroll_result_message (self, ENROLL_STATE_WARNING, message);
         }
@@ -744,6 +823,7 @@ enroll_start_cb (GObject      *object,
                  gpointer      user_data)
 {
   g_autoptr(GError) error = NULL;
+  g_autoptr(DialogStateRemover) state_remover = NULL;
   CcFprintdDevice *fprintd_device = CC_FPRINTD_DEVICE (object);
   CcFingerprintDialog *self = user_data;
 
@@ -752,17 +832,16 @@ enroll_start_cb (GObject      *object,
   if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
     return;
 
-  gtk_spinner_stop (self->spinner);
+  state_remover = auto_state_remover (self, DIALOG_STATE_DEVICE_ENROLL_STARTING);
 
   if (error)
     {
       g_autofree char *error_message = NULL;
 
-      self->enrolling_finger = NULL;
+      remove_dialog_state (self, DIALOG_STATE_DEVICE_ENROLLING);
 
-      g_dbus_error_strip_remote_error (error);
       error_message = g_strdup_printf (_("Failed to start enrollment: %s"),
-                                       error->message);
+                                       dbus_error_to_human (self, error));
       g_warning ("Enrollment on device %s failed: %s",
                  cc_fprintd_device_get_name (self->device), error->message);
       notify_error (self, error_message);
@@ -782,6 +861,7 @@ enroll_stop_cb (GObject      *object,
                 gpointer      user_data)
 {
   g_autoptr(GError) error = NULL;
+  g_autoptr(DialogStateRemover) state_remover = NULL;
   CcFprintdDevice *fprintd_device = CC_FPRINTD_DEVICE (object);
   CcFingerprintDialog *self = user_data;
 
@@ -790,8 +870,8 @@ enroll_stop_cb (GObject      *object,
   if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
     return;
 
-  self->enrolling_finger = NULL;
-  gtk_spinner_stop (self->spinner);
+  state_remover = auto_state_remover (self, DIALOG_STATE_DEVICE_ENROLLING |
+                                            DIALOG_STATE_DEVICE_ENROLL_STOPPING);
   gtk_widget_set_sensitive (self->enrollment_view, TRUE);
   gtk_stack_set_visible_child (self->stack, self->prints_manager);
 
@@ -799,9 +879,8 @@ enroll_stop_cb (GObject      *object,
     {
       g_autofree char *error_message = NULL;
 
-      g_dbus_error_strip_remote_error (error);
       error_message = g_strdup_printf (_("Failed to stop enrollment: %s"),
-                                       error->message);
+                                       dbus_error_to_human (self, error));
       g_warning ("Stopping enrollment on device %s failed: %s",
                  cc_fprintd_device_get_name (self->device), error->message);
       notify_error (self, error_message);
@@ -815,9 +894,11 @@ enroll_stop_cb (GObject      *object,
 static void
 enroll_stop (CcFingerprintDialog *self)
 {
-  g_return_if_fail (self->enrolling_finger);
+  g_return_if_fail (self->dialog_state & DIALOG_STATE_DEVICE_ENROLLING);
+
+  if (!add_dialog_state (self, DIALOG_STATE_DEVICE_ENROLL_STOPPING))
+    return;
 
-  gtk_spinner_start (self->spinner);
   gtk_widget_set_sensitive (self->enrollment_view, FALSE);
   cc_fprintd_device_call_enroll_stop (self->device, self->cancellable,
                                       enroll_stop_cb, self);
@@ -857,7 +938,10 @@ enroll_finger (CcFingerprintDialog *self,
 
   g_return_if_fail (finger_id);
 
-  self->enrolling_finger = finger_id;
+  if (!add_dialog_state (self, DIALOG_STATE_DEVICE_ENROLLING |
+                               DIALOG_STATE_DEVICE_ENROLL_STARTING))
+    return;
+
   self->enroll_progress = 0;
   self->enroll_stages_passed = 0;
 
@@ -870,8 +954,7 @@ enroll_finger (CcFingerprintDialog *self,
   set_enroll_result_message (self, ENROLL_STATE_NORMAL, NULL);
   gtk_stack_set_visible_child (self->stack, self->enrollment_view);
   gtk_label_set_label (self->enroll_message, enroll_message);
-  gtk_entry_set_text (self->enroll_print_entry, finger_name);
-  gtk_spinner_start (self->spinner);
+  gtk_editable_set_text (GTK_EDITABLE (self->enroll_print_entry), finger_name);
 
   cc_fprintd_device_call_enroll_start (self->device, finger_id, self->cancellable,
                                        enroll_start_cb, self);
@@ -890,26 +973,10 @@ populate_enrollment_view (CcFingerprintDialog *self)
                           (GtkWidget **) &self->enroll_result_image,
                           (GtkWidget **) &self->enroll_result_message);
 
-  gtk_container_add (GTK_CONTAINER (self->enroll_print_bin), self->enroll_result_icon);
+  gtk_box_prepend (GTK_BOX (self->enroll_print_bin), self->enroll_result_icon);
 
   style_context = gtk_widget_get_style_context (self->enroll_result_icon);
   gtk_style_context_add_class (style_context,  "enroll-status");
-
-  gtk_widget_show_all (self->enroll_print_bin);
-}
-
-static void
-reenroll_finger_cb (CcFingerprintDialog *self)
-{
-  GtkWidget *button;
-  GtkWidget *flowbox_child;
-  const char *finger_id;
-
-  button = gtk_popover_get_relative_to (GTK_POPOVER (self->print_popover));
-  flowbox_child = g_object_get_data (G_OBJECT (button), "flowbox-child");
-  finger_id = g_object_get_data (G_OBJECT (flowbox_child), "finger-id");
-
-  enroll_finger (self, finger_id);
 }
 
 static void
@@ -920,12 +987,12 @@ on_print_activated_cb (GtkFlowBox          *flowbox,
   GtkWidget *selected_button;
 
   selected_button = g_object_get_data (G_OBJECT (child), "button");
-  gtk_button_clicked (GTK_BUTTON (selected_button));
+  g_signal_emit_by_name (GTK_MENU_BUTTON (selected_button), "activate");
 }
 
 static void
 on_enroll_cb (CcFingerprintDialog *self,
-              GtkModelButton      *button)
+              GtkMenuButton       *button)
 {
   const char *finger_id;
 
@@ -942,11 +1009,11 @@ populate_add_print_popover (CcFingerprintDialog *self)
     {
       GtkWidget *finger_item;
 
-      finger_item = gtk_model_button_new ();
+      finger_item = gtk_button_new ();
       gtk_button_set_label (GTK_BUTTON (finger_item), get_finger_name (FINGER_IDS[i]));
       gtk_button_set_use_underline (GTK_BUTTON (finger_item), TRUE);
       g_object_set_data (G_OBJECT (finger_item), "finger-id", (gpointer) FINGER_IDS[i]);
-      gtk_container_add (self->add_print_popover_box, finger_item);
+      gtk_box_prepend (GTK_BOX (self->add_print_popover_box), finger_item);
 
       g_signal_connect_object (finger_item, "clicked", G_CALLBACK (on_enroll_cb),
                                self, G_CONNECT_SWAPPED);
@@ -966,19 +1033,25 @@ populate_prints_gallery (CcFingerprintDialog *self)
   for (i = 0; i < N_VALID_FINGERS; ++i)
     {
       GtkWidget *flowbox_child;
+      GtkWidget *popover;
+      GtkWidget *reenroll_button;
 
       flowbox_child = fingerprint_menu_button ("fingerprint-detection-symbolic",
                                                get_finger_name (FINGER_IDS[i]));
 
       button = g_object_get_data (G_OBJECT (flowbox_child), "button");
 
-      gtk_menu_button_set_popover (GTK_MENU_BUTTON (button),
-                                   GTK_WIDGET (self->print_popover));
-      /* Move the popover on click, so we can just reuse the same instance */
-      g_signal_connect_object (button, "clicked",
-                               G_CALLBACK (gtk_popover_set_relative_to),
-                               self->print_popover, G_CONNECT_SWAPPED);
+      popover = gtk_popover_new ();
+      reenroll_button = gtk_button_new ();
+      gtk_button_set_use_underline (GTK_BUTTON (reenroll_button), TRUE);
+      gtk_button_set_label (GTK_BUTTON (reenroll_button), _("_Re-enroll this finger…"));
+      g_object_set_data (G_OBJECT (reenroll_button), "finger-id",
+                         (gpointer) FINGER_IDS[i]);
+      g_signal_connect_object (reenroll_button, "clicked", G_CALLBACK (on_enroll_cb), self, G_CONNECT_SWAPPED);
+      gtk_popover_set_child (GTK_POPOVER (popover), reenroll_button);
 
+      gtk_menu_button_set_popover (GTK_MENU_BUTTON (button),
+                                   popover);
       g_object_set_data (G_OBJECT (flowbox_child), "finger-id",
                          (gpointer) FINGER_IDS[i]);
 
@@ -1000,7 +1073,6 @@ populate_prints_gallery (CcFingerprintDialog *self)
   gtk_flow_box_insert (self->prints_gallery, self->add_print_icon, -1);
   gtk_flow_box_set_max_children_per_line (self->prints_gallery, 1);
 
-  gtk_widget_show_all (GTK_WIDGET (self->prints_gallery));
   gtk_flow_box_set_filter_func (self->prints_gallery, prints_visibility_filter,
                                 self, NULL);
 
@@ -1025,23 +1097,23 @@ release_device_cb (GObject      *object,
     {
       g_autofree char *error_message = NULL;
 
-      g_dbus_error_strip_remote_error (error);
       error_message = g_strdup_printf (_("Failed to release fingerprint device %s: %s"),
                                        cc_fprintd_device_get_name (fprintd_device),
-                                       error->message);
-      g_warning ("%s", error_message);
+                                       dbus_error_to_human (self, error));
+      g_warning ("Releasing device %s failed: %s",
+                 cc_fprintd_device_get_name (self->device), error->message);
 
       notify_error (self, error_message);
       return;
     }
 
-  self->device_claimed = FALSE;
+  remove_dialog_state (self, DIALOG_STATE_DEVICE_CLAIMED);
 }
 
 static void
 release_device (CcFingerprintDialog *self)
 {
-  if (!self->device || !self->device_claimed)
+  if (!self->device || !(self->dialog_state & DIALOG_STATE_DEVICE_CLAIMED))
     return;
 
   disconnect_device_signals (self);
@@ -1089,19 +1161,18 @@ on_device_owner_changed (CcFprintdDevice     *device,
 
   if (!name_owner)
     {
-      if (self->device_claimed)
+      if (self->dialog_state & DIALOG_STATE_DEVICE_CLAIMED)
         {
           disconnect_device_signals (self);
 
-          if (self->enrolling_finger)
+          if (self->dialog_state & DIALOG_STATE_DEVICE_ENROLLING)
             {
               set_enroll_result_message (self, ENROLL_STATE_ERROR,
                                          C_("Fingerprint enroll state",
                                             "Problem Reading Device"));
-              self->enrolling_finger = NULL;
             }
 
-          self->device_claimed = FALSE;
+          remove_dialog_state (self, DIALOG_STATE_DEVICE_CLAIMED);
           claim_device (self);
         }
     }
@@ -1113,6 +1184,7 @@ claim_device_cb (GObject      *object,
                  gpointer      user_data)
 {
   g_autoptr(GError) error = NULL;
+  g_autoptr(DialogStateRemover) state_remover = NULL;
   CcFprintdDevice *fprintd_device = CC_FPRINTD_DEVICE (object);
   CcFingerprintDialog *self = user_data;
 
@@ -1121,28 +1193,30 @@ claim_device_cb (GObject      *object,
   if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
     return;
 
-  self->claiming = FALSE;
+  state_remover = auto_state_remover (self, DIALOG_STATE_DEVICE_CLAIMING);
 
   if (error)
     {
       g_autofree char *dbus_error = g_dbus_error_get_remote_error (error);
       g_autofree char *error_message = NULL;
 
-      if (dbus_error && g_str_has_suffix (dbus_error, ".Error.AlreadyInUse") &&
-          self->device_claimed)
+      if (dbus_error && g_str_equal (dbus_error, CC_FPRINTD_NAME ".Error.AlreadyInUse") &&
+          (self->dialog_state & DIALOG_STATE_DEVICE_CLAIMED))
          return;
 
-      g_dbus_error_strip_remote_error (error);
       error_message = g_strdup_printf (_("Failed to claim fingerprint device %s: %s"),
                                        cc_fprintd_device_get_name (self->device),
-                                       error->message);
-      g_warning ("%s", error_message);
+                                       dbus_error_to_human (self, error));
+      g_warning ("Claiming device %s failed: %s",
+                 cc_fprintd_device_get_name (self->device), error->message);
       notify_error (self, error_message);
       return;
     }
 
+  if (!add_dialog_state (self, DIALOG_STATE_DEVICE_CLAIMED))
+    return;
+
   gtk_widget_set_sensitive (self->prints_manager, TRUE);
-  self->device_claimed = TRUE;
   self->device_signal_id = g_signal_connect_object (self->device, "g-signal",
                                                     G_CALLBACK (on_device_signal),
                                                     self, G_CONNECT_SWAPPED);
@@ -1156,14 +1230,13 @@ claim_device (CcFingerprintDialog *self)
 {
   ActUser *user;
 
-  g_return_if_fail (!self->device_claimed);
+  g_return_if_fail (!(self->dialog_state & DIALOG_STATE_DEVICE_CLAIMED));
 
-  if (self->claiming)
+  if (!add_dialog_state (self, DIALOG_STATE_DEVICE_CLAIMING))
     return;
 
   user = cc_fingerprint_manager_get_user (self->manager);
   gtk_widget_set_sensitive (self->prints_manager, FALSE);
-  self->claiming = TRUE;
 
   cc_fprintd_device_call_claim (self->device,
                                 act_user_get_user_name (user),
@@ -1184,7 +1257,8 @@ on_stack_child_changed (CcFingerprintDialog *self)
   gtk_widget_hide (GTK_WIDGET (self->cancel_button));
   gtk_widget_hide (GTK_WIDGET (self->done_button));
 
-  gtk_header_bar_set_show_close_button (self->titlebar, TRUE);
+  adw_header_bar_set_show_start_title_buttons (ADW_HEADER_BAR (self->titlebar), TRUE);
+  adw_header_bar_set_show_end_title_buttons (ADW_HEADER_BAR (self->titlebar), TRUE);
   gtk_flow_box_invalidate_filter (self->prints_gallery);
 
   if (visible_child == self->prints_manager)
@@ -1194,12 +1268,13 @@ on_stack_child_changed (CcFingerprintDialog *self)
       notify_error (self, NULL);
       update_prints_store (self);
 
-      if (!self->device_claimed)
+      if (!(self->dialog_state & DIALOG_STATE_DEVICE_CLAIMED))
         claim_device (self);
     }
   else if (visible_child == self->enrollment_view)
     {
-      gtk_header_bar_set_show_close_button (self->titlebar, FALSE);
+      adw_header_bar_set_show_start_title_buttons (ADW_HEADER_BAR (self->titlebar), FALSE);
+      adw_header_bar_set_show_end_title_buttons (ADW_HEADER_BAR (self->titlebar), FALSE);
 
       gtk_widget_show (GTK_WIDGET (self->cancel_button));
       gtk_widget_set_sensitive (GTK_WIDGET (self->cancel_button), TRUE);
@@ -1226,9 +1301,9 @@ cc_fingerprint_dialog_init (CcFingerprintDialog *self)
   provider = gtk_css_provider_new ();
   gtk_css_provider_load_from_resource (provider,
                                        "/org/gnome/control-center/user-accounts/cc-fingerprint-dialog.css");
-  gtk_style_context_add_provider_for_screen (gdk_screen_get_default (),
-                                             GTK_STYLE_PROVIDER (provider),
-                                             GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
+  gtk_style_context_add_provider_for_display (gdk_display_get_default (),
+                                              GTK_STYLE_PROVIDER (provider),
+                                              GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
 
   on_stack_child_changed (self);
   g_signal_connect_object (self->stack, "notify::visible-child",
@@ -1261,13 +1336,18 @@ on_devices_list (GObject      *object,
                  gpointer      user_data)
 {
   g_autolist (CcFprintdDevice) fprintd_devices = NULL;
+  g_autoptr(DialogStateRemover) state_remover = NULL;
   g_autoptr(GError) error = NULL;
   CcFingerprintManager *fingerprint_manager = CC_FINGERPRINT_MANAGER (object);
   CcFingerprintDialog *self = CC_FINGERPRINT_DIALOG (user_data);
 
   fprintd_devices = cc_fingerprint_manager_get_devices_finish (fingerprint_manager,
                                                                res, &error);
-  gtk_spinner_stop (self->spinner);
+
+  if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+    return;
+
+  state_remover = auto_state_remover (self, DIALOG_STATE_DEVICES_LISTING);
 
   if (fprintd_devices == NULL)
     {
@@ -1275,10 +1355,9 @@ on_devices_list (GObject      *object,
         {
           g_autofree char *error_message = NULL;
 
-          g_dbus_error_strip_remote_error (error);
           error_message = g_strdup_printf (_("Failed to get fingerprint devices: %s"),
-                                           error->message);
-          g_warning ("%s", error_message);
+                                           dbus_error_to_human (self, error));
+          g_warning ("Retrieving fingerprint devices failed: %s", error->message);
           notify_error (self, error_message);
         }
 
@@ -1322,7 +1401,7 @@ cc_fingerprint_dialog_constructed (GObject *object)
   bindtextdomain ("fprintd", GNOMELOCALEDIR);
   bind_textdomain_codeset ("fprintd", "UTF-8");
 
-  gtk_spinner_start (self->spinner);
+  add_dialog_state (self, DIALOG_STATE_DEVICES_LISTING);
   cc_fingerprint_manager_get_devices (self->manager, self->cancellable,
                                       on_devices_list, self);
 }
@@ -1364,7 +1443,7 @@ delete_prints_button_clicked_cb (CcFingerprintDialog *self)
 static void
 cancel_button_clicked_cb (CcFingerprintDialog *self)
 {
-  if (self->enrolling_finger)
+  if (self->dialog_state & DIALOG_STATE_DEVICE_ENROLLING)
     {
       g_cancellable_cancel (self->cancellable);
       g_set_object (&self->cancellable, g_cancellable_new ());
@@ -1381,7 +1460,7 @@ cancel_button_clicked_cb (CcFingerprintDialog *self)
 static void
 done_button_clicked_cb (CcFingerprintDialog *self)
 {
-  g_return_if_fail (self->enrolling_finger);
+  g_return_if_fail (self->dialog_state & DIALOG_STATE_DEVICE_ENROLLING);
 
   g_debug ("Completing enroll operation");
   enroll_stop (self);
@@ -1391,7 +1470,7 @@ static void
 fingerprint_dialog_delete_cb (CcFingerprintDialog *self)
 {
   cc_fingerprint_manager_update_state (self->manager, NULL, NULL);
-  gtk_widget_destroy (GTK_WIDGET (self));
+  gtk_window_destroy (GTK_WINDOW (self));
 }
 
 static void
@@ -1433,7 +1512,6 @@ cc_fingerprint_dialog_class_init (CcFingerprintDialogClass *klass)
   gtk_widget_class_bind_template_child (widget_class, CcFingerprintDialog, error_infobar);
   gtk_widget_class_bind_template_child (widget_class, CcFingerprintDialog, infobar_error);
   gtk_widget_class_bind_template_child (widget_class, CcFingerprintDialog, no_devices_found);
-  gtk_widget_class_bind_template_child (widget_class, CcFingerprintDialog, print_popover);
   gtk_widget_class_bind_template_child (widget_class, CcFingerprintDialog, prints_gallery);
   gtk_widget_class_bind_template_child (widget_class, CcFingerprintDialog, prints_manager);
   gtk_widget_class_bind_template_child (widget_class, CcFingerprintDialog, spinner);
@@ -1449,6 +1527,5 @@ cc_fingerprint_dialog_class_init (CcFingerprintDialogClass *klass)
   gtk_widget_class_bind_template_callback (widget_class, done_button_clicked_cb);
   gtk_widget_class_bind_template_callback (widget_class, fingerprint_dialog_delete_cb);
   gtk_widget_class_bind_template_callback (widget_class, on_print_activated_cb);
-  gtk_widget_class_bind_template_callback (widget_class, reenroll_finger_cb);
   gtk_widget_class_bind_template_callback (widget_class, select_device_row);
 }
